__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import logging
import logging.handlers
import os
import platform
import re
import sys
from typing import Optional

from pkg_resources import resource_filename

from . import formatter
from ..enums import LogVerbosity
from ..jaml import JAML


class NTLogger:
    """A compatible logger for Windows system, colors are all removed to keep compatible."""

    def __init__(self, context: str, log_level: 'LogVerbosity' = LogVerbosity.INFO):
        """
        Create a compatible logger for Windows system, colors are all removed to keep compatible.

        :param context: The name prefix of each log.
        :param log_level: Level of log.
        """
        self.context = self._planify(context)
        self.log_level = log_level

    @staticmethod
    def _planify(msg):
        return re.sub(r'\u001b\[.*?[@-~]', '', msg)

    def info(self, msg: str, **kwargs):
        """
        Log info-level message.

        :param msg: Context of log.
        """
        if self.log_level <= LogVerbosity.INFO:
            sys.stdout.write(f'{self.context}[I]:{self._planify(msg)}')

    def critical(self, msg: str, **kwargs):
        """
        Log critical-level message.

        :param msg: Context of log.
        """
        if self.log_level <= LogVerbosity.CRITICAL:
            sys.stdout.write(f'{self.context}[C]:{self._planify(msg)}')

    def debug(self, msg: str, **kwargs):
        """
        Log debug-level message.

        :param msg: Content of log.
        """
        if self.log_level <= LogVerbosity.DEBUG:
            sys.stdout.write(f'{self.context}[D]:{self._planify(msg)}')

    def error(self, msg: str, **kwargs):
        """
        Log error-level message.

        :param msg: Context of log.
        """
        if self.log_level <= LogVerbosity.ERROR:
            sys.stdout.write(f'{self.context}[E]:{self._planify(msg)}')

    def warning(self, msg: str, **kwargs):
        """
        Log warning-level message.

        :param msg: Context of log.
        """
        if self.log_level <= LogVerbosity.WARNING:
            sys.stdout.write(f'{self.context}[W]:{self._planify(msg)}')

    def success(self, msg: str, **kwargs):
        """
        Log success-level message.

        :param msg: Context of log.
        """
        if self.log_level <= LogVerbosity.SUCCESS:
            sys.stdout.write(f'{self.context}[S]:{self._planify(msg)}')


class PrintLogger(NTLogger):

    @staticmethod
    def _planify(msg):
        return msg


class SysLogHandlerWrapper(logging.handlers.SysLogHandler):
    """
    Override the priority_map :class:`SysLogHandler`.

    .. warning::
        This messages at DEBUG and INFO are therefore not stored by ASL, (ASL = Apple System Log)
        which in turn means they can't be printed by syslog after the fact. You can confirm it via :command:`syslog` or
        :command:`tail -f /var/log/system.log`.
    """

    priority_map = {
        'DEBUG': 'debug',
        'INFO': 'info',
        'WARNING': 'warning',
        'ERROR': 'error',
        'CRITICAL': 'critical',
        'SUCCESS': 'notice'
    }


class JinaLogger:
    supported = {'FileHandler', 'StreamHandler', 'SysLogHandler', 'FluentHandler'}

    def __init__(self,
                 context: str,
                 name: Optional[str] = None,
                 log_config: Optional[str] = None,
                 identity: Optional[str] = None,
                 workspace_path: Optional[str] = None,
                 quiet: bool = False,
                 **kwargs):
        """
        Build a logger for a context.

        :param context: The context identifier of the class, module or method.
        :param log_config: The configuration file for the logger.
        :param identity: The id of the group the messages from this logger will belong, used by fluentd default
        configuration to group logs by pod.
        :param workspace_path: The workspace path where the log will be stored at (only apply to fluentd)
        :returns: an executor object.
        """
        from .. import __uptime__
        if not log_config:
            log_config = os.getenv('JINA_LOG_CONFIG',
                                   resource_filename('jina', '/'.join(
                                       ('resources', 'logging.default.yml'))))

        if quiet or os.getenv('JINA_LOG_CONFIG', None) == 'QUIET':
            log_config = resource_filename('jina', '/'.join(
                ('resources', 'logging.quiet.yml')))

        if not identity:
            identity = os.getenv('JINA_LOG_ID', None)

        if not name:
            name = os.getenv('JINA_POD_NAME', context)

        # Remove all handlers associated with the root logger object.
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

        self.logger = logging.getLogger(context)
        self.logger.propagate = False

        if workspace_path is None:
            workspace_path = os.getenv('JINA_LOG_WORKSPACE', '/tmp/jina/')

        context_vars = {'name': name,
                        'uptime': __uptime__,
                        'context': context,
                        'workspace_path': workspace_path}
        if identity:
            context_vars['log_id'] = identity

        self.add_handlers(log_config, **context_vars)

        # note logger.success isn't default there
        success_level = LogVerbosity.SUCCESS.value  # between WARNING and INFO
        logging.addLevelName(success_level, 'SUCCESS')
        setattr(self.logger, 'success', lambda message: self.logger.log(success_level, message))

        self.info = self.logger.info
        self.critical = self.logger.critical
        self.debug = self.logger.debug
        self.error = self.logger.error
        self.warning = self.logger.warning
        self.success = self.logger.success

    @property
    def handlers(self):
        """
        Get the handlers of the logger.

        :returns: Handlers of logger.
        """
        return self.logger.handlers

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """
        Close all the handlers.

        :returns: None
        """
        for handler in self.logger.handlers:
            handler.close()

    def add_handlers(self, config_path: str = None, **kwargs):
        """
        Add handlers from config file.

        :param config_path: Path of config file.
        :param kwargs: Extra parameters.
        :returns: None
        """
        self.logger.handlers = []

        with open(config_path) as fp:
            config = JAML.load(fp)

        for h in config['handlers']:
            cfg = config['configs'].get(h, None)
            fmt = getattr(formatter, cfg.get('formatter', 'PlainFormatter'))

            if h not in self.supported or not cfg:
                raise ValueError(f'can not find configs for {h}, maybe it is not supported')

            handler = None
            if h == 'StreamHandler':
                handler = logging.StreamHandler(sys.stdout)
                handler.setFormatter(fmt(cfg['format'].format_map(kwargs)))
            elif h == 'SysLogHandler':
                if cfg['host'] and cfg['port']:
                    handler = SysLogHandlerWrapper(address=(cfg['host'], cfg['port']))
                else:
                    # a UNIX socket is used
                    if platform.system() == 'Darwin':
                        handler = SysLogHandlerWrapper(address='/var/run/syslog')
                    else:
                        handler = SysLogHandlerWrapper(address='/dev/log')
                if handler:
                    handler.ident = cfg.get('ident', '')
                    handler.setFormatter(fmt(cfg['format'].format_map(kwargs)))

                try:
                    handler._connect_unixsocket(handler.address)
                except OSError:
                    handler = None
                    pass
            elif h == 'FileHandler':
                handler = logging.FileHandler(cfg['output'].format_map(kwargs), delay=True)
                handler.setFormatter(fmt(cfg['format'].format_map(kwargs)))
            elif h == 'FluentHandler':
                from ..importer import ImportExtensions
                with ImportExtensions(required=False, verbose=False):
                    from fluent import asynchandler as fluentasynchandler
                    from fluent.handler import FluentRecordFormatter

                    handler = fluentasynchandler.FluentHandler(cfg['tag'],
                                                               host=cfg['host'],
                                                               port=cfg['port'], queue_circular=True)

                    cfg['format'].update(kwargs)
                    fmt = FluentRecordFormatter(cfg['format'])
                    handler.setFormatter(fmt)

            if handler:
                self.logger.addHandler(handler)

        verbose_level = LogVerbosity.from_string(config['level'])
        if 'JINA_LOG_LEVEL' in os.environ:
            verbose_level = LogVerbosity.from_string(os.environ['JINA_LOG_LEVEL'])
        self.logger.setLevel(verbose_level.value)



from typing import Dict, List, Union

import numpy as np

from jina.executors.crafters import BaseCrafter


class AlbumentationsCrafter(BaseCrafter):
    """Apply transforms from the Albumentations package to the image.

    This crafter allows you to apply any of the Albumenation's transforms to
    an image. You can also compose as many transforms as you would like inside
    a single object. For a full list of available transforms, visit
    `Albumentations's GitHub page <https://github.com/albumentations-team/albumentations/#list-of-augmentations>`_
    or see `the documentation <https://albumentations.ai/docs/api_reference/augmentations/transforms/>`_.

    .. attention::
        Albumentations' transforms were created for the purpose of image augmentation,
        where the transforms are applied randomly. However in indexing/search
        context you will want the transforms to be applied deterministically.

        For that reason pay attention to the parameters and set them in a way that
        any randomness is removed. For example, if you want to rotate the image 45
        degrees, set the ``limit`` parameter in the ``Rotate`` transformation to
        ``[45,45]``. And you don't want to use some inherently random transforms,
        such as ``RandomRotate90``.

        We take care of some of this automatically: we set ``always_apply=True`` for
        all transforms, so that they are always applied.

    Example:
        The transforms that you provide will be applied in the order provided using
        ``A.Compose``. For example, the ``crafter`` in the example below

        .. code-block:: python

            from jina.hub.crafters.image.AlbumentationsCrafter import AlbumentationsCrafter

            transforms = [
                'HorizontalFlip',
                {'Rotate': {'limit': [45, 45]}}
            ]
            crafter = AlbumentationsCrafter(transforms)

        will have the equivalent effect as the ``transform`` below

        .. code-block:: python

            import albumentations as A

            transform = A.Compose([
                A.HorizontalFlip(always_apply=True),
                A.Rotate(limit=[45,45], always_apply=True)
            ])

    :param transforms: A list of transformations that should be applied.
        Each item in the list should be of the form
        ``{'TransformClass': kwargs}`` or ``'TransformClass'``.
    :param args:  Additional positional arguments
    :param kwargs: A dictionary with keyword arguments that will
        be passed to ``TransformClass`` at initialization
        (or an empty dict if there are no such arguments).
        The argument ``always_apply=True`` will be added to all kwargs
        automatically.
    """

    def __init__(
        self,
        transforms: List[Union[str, Dict[str, Dict]]] = ['HorizontalFlip'],
        *args,
        **kwargs
    ):
        """Set constructor."""
        super().__init__(*args, **kwargs)

        if not isinstance(transforms, list):
            self.logger.error('The `transform` argument has to be a list.')
            raise ValueError

        self.transforms_specification = transforms

    def post_init(self):
        import albumentations as A

        transforms_list = []
        for transform in self.transforms_specification:

            if isinstance(transform, str):
                tr_name, tr_kwargs = transform, {}
            elif isinstance(transform, dict):
                tr_name, tr_kwargs = next(iter(transform.items()))
            else:
                self.logger.error(
                    'Each element of the `transform` has to be either a dict'
                    ' or a string.'
                )
                raise ValueError

            tr_kwargs.update({'always_apply': True})

            try:
                tr_class = getattr(A, tr_name)
            except:
                self.logger.error(
                    f'The albumentations class `{tr_name}` does not seem to exist,'
                    ' check that you have not made a typo.'
                )
                raise ValueError

            try:
                alb_transform = tr_class(**tr_kwargs)
            except:
                self.logger.error(
                    f'Error instantiating albumentations class `{tr_name}`,'
                    ' check that all kwargs are valid.'
                )
                raise ValueError

            transforms_list.append(alb_transform)

        self.transforms = A.Compose(transforms_list)

    def craft(self, blob: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """Apply transformations to the image.

        :param blob: The image to transform, should be in ``[H, W, C]`` format,
            where ``C`` is the color channel, which is either RGB (siz 3), or BW
            (size 1).
        :return: The transformed image
        """

        return self.transforms(image=blob)['image']




from os import path
from typing import Optional, Iterable

import numpy as np

from jina.executors.decorators import as_update_method
from jina.executors.indexers.vector import NumpyIndexer
from jina.helper import cached_property

if False:
    import zarr


class ZarrIndexer(NumpyIndexer):
    """
    Indexing based on Zarr arrays.

    For more information about Zarr, please check
    https://zarr.readthedocs.io/en/stable/index.html

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_add_handler(self) -> 'zarr.hierarchy.Group':
        """
        Open an existing zarr group file for adding new vectors.

        :return: a zarr group file
        """
        import zarr
        return zarr.open(store=self.index_abspath, mode='a')

    def get_create_handler(self) -> 'zarr.hierarchy.Group':
        """
        Create a zarr group file for adding new vectors.

        :return: a zarr group file
        """
        import zarr
        return zarr.open(store=self.index_abspath, mode='w')

    @as_update_method
    def add(self, keys: Iterable[str], vectors: 'np.ndarray', *args, **kwargs) -> None:
        """
        Add the embeddings and document ids to the index.

        :param keys: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param vectors: embeddings
        """
        np_keys = np.array(keys, (np.str_, self.key_length))

        self._add(np_keys, vectors)

    def _add(self, keys: 'np.ndarray', vectors: 'np.ndarray'):
        self._validate_key_vector_shapes(keys, vectors)
        if 'default' in self.write_handler.array_keys():
            self.write_handler['default'].append(data=vectors)
        else:
            self.write_handler.array(name='default', data=vectors)
        self.valid_indices = np.concatenate((self.valid_indices, np.full(len(keys), True)))
        self.key_bytes += keys.tobytes()
        self._size += keys.shape[0]

    @property
    def query_handler(self) -> Optional['zarr.core.Array']:
        """
        Get zarr file handler.

        :return: Zarr query handler
        """
        return self.get_query_handler()

    def get_query_handler(self) -> Optional['zarr.core.Array']:
        """
        Get zarr file handler.

        :return: Zarr query handler
        """
        import zarr
        if not (path.exists(self.index_abspath) or self.num_dim or self.dtype):
            return
        return zarr.open(store=f'{self.index_abspath}/default', mode='r',
                         shape=(self._size, self.num_dim), chunks=True)

    def query_by_key(self, keys: Iterable[str], *args, **kwargs) -> 'np.ndarray':
        """
        Get the vectors by keys.

        :param keys: list of document keys` as 1D-ndarray
        :return: subset of indexed vectors
        """
        filtered_keys = self._filter_nonexistent_keys(keys, self._ext2int_id.keys())
        int_ids = [self._ext2int_id[j] for j in filtered_keys]
        return self._raw_ndarray.get_orthogonal_selection(int_ids)

    @cached_property
    def _raw_ndarray(self):
        a = self.query_handler
        return a



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTorchEncoder


class LaserEncoder(BaseTorchEncoder):
    """
    Encode an array of string in size `B` into an ndarray in size `B x D`

    The ndarray potentially is BatchSize x (Channel x Height x Width)

    :class:`LaserEncoder` is a encoder based on Facebook Research's LASER
    (Language-Agnostic SEntence Representations) to compute multilingual
    sentence embeddings: https://github.com/facebookresearch/LASER
    :param path_to_bpe_codes: path to bpe codes from Laser.
        Defaults to Laser.DEFAULT_BPE_CODES_FILE.
    :param path_to_bpe_vocab: path to bpe vocabs from Laser.
        Defaults to Laser.DEFAULT_BPE_VOCAB_FILE.
    :param path_to_encoder: path to the encoder from Laser.
        Defaults to Laser.DEFAULT_ENCODER_FILE.
    :param language: language of the text. Defaults to english(en).
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(
            self,
            path_to_bpe_codes: str = None,
            path_to_bpe_vocab: str = None,
            path_to_encoder: str = None,
            language: str = 'en',
            *args,
            **kwargs,
    ):
        """Set Constructor."""
        super().__init__(*args, **kwargs)
        from laserembeddings import Laser
        self._path_to_bpe_codes = path_to_bpe_codes or Laser.DEFAULT_BPE_CODES_FILE
        self._path_to_bpe_vocab = path_to_bpe_vocab or Laser.DEFAULT_BPE_VOCAB_FILE
        self._path_to_encoder = path_to_encoder or Laser.DEFAULT_ENCODER_FILE
        self.language = language.lower()

    def post_init(self):
        from laserembeddings import Laser
        self.model = Laser(
            bpe_codes=self._path_to_bpe_codes,
            bpe_vocab=self._path_to_bpe_vocab,
            encoder=self._path_to_encoder,
        )
        self.to_device(self.model.bpeSentenceEmbedding.encoder.encoder)

    @batching
    @as_ndarray
    def encode(self, data: "np.ndarray", *args, **kwargs) -> "np.ndarray":
        """
         Encode data into an ndarray in size `B x D`.

         B is the `Batch size` and `D` the dimension.

        :param data: a 1d array of string type in size `B`
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        :return: an ndarray in size `B x D`.
        """
        return self.model.embed_sentences(data, lang=self.language)



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

# Copyright 2018 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Class for text data."""
import string
import numpy as np
import torch


class SimpleVocab(object):

  def __init__(self):
    super(SimpleVocab, self).__init__()
    self.word2id = {}
    self.wordcount = {}
    self.word2id['<UNK>'] = 0
    self.wordcount['<UNK>'] = 9e9

  def tokenize_text(self, text):
    text = text.encode('ascii', 'ignore').decode('ascii')
    tokens = str(text).lower().translate(str.maketrans("","",
                                         string.punctuation)).strip().split()
    return tokens

  def add_text_to_vocab(self, text):
    tokens = self.tokenize_text(text)
    for token in tokens:
      if token not in self.word2id:
        self.word2id[token] = len(self.word2id)
        self.wordcount[token] = 0
      self.wordcount[token] += 1

  def threshold_rare_words(self, wordcount_threshold=5):
    for w in self.word2id:
      if self.wordcount[w] < wordcount_threshold:
        self.word2id[w] = 0

  def encode_text(self, text):
    tokens = self.tokenize_text(text)
    x = [self.word2id.get(t, 0) for t in tokens]
    return x

  def get_size(self):
    return len(self.word2id)


class TextLSTMModel(torch.nn.Module):

  def __init__(self,
               texts_to_build_vocab,
               word_embed_dim=512,
               lstm_hidden_dim=512):

    super(TextLSTMModel, self).__init__()

    self.vocab = SimpleVocab()
    for text in texts_to_build_vocab:
      self.vocab.add_text_to_vocab(text)
    vocab_size = self.vocab.get_size()

    self.word_embed_dim = word_embed_dim
    self.lstm_hidden_dim = lstm_hidden_dim
    self.embedding_layer = torch.nn.Embedding(vocab_size, word_embed_dim)
    self.lstm = torch.nn.LSTM(word_embed_dim, lstm_hidden_dim)
    self.fc_output = torch.nn.Sequential(
        torch.nn.Dropout(p=0.1),
        torch.nn.Linear(lstm_hidden_dim, lstm_hidden_dim),
    )

  def forward(self, x):
    """ input x: list of strings"""
    if type(x) is list:
      if type(x[0]) is str or type(x[0]) is unicode:
        x = [self.vocab.encode_text(text) for text in x]

    assert type(x) is list
    assert type(x[0]) is list
    assert type(x[0][0]) is int
    return self.forward_encoded_texts(x)

  def forward_encoded_texts(self, texts):
    # to tensor
    lengths = [len(t) for t in texts]
    itexts = torch.zeros((np.max(lengths), len(texts))).long()
    for i in range(len(texts)):
      itexts[:lengths[i], i] = torch.tensor(texts[i])

    # embed words
    itexts = torch.autograd.Variable(itexts)
    etexts = self.embedding_layer(itexts)

    # lstm
    lstm_output, _ = self.forward_lstm_(etexts)

    # get last output (using length)
    text_features = []
    for i in range(len(texts)):
      text_features.append(lstm_output[lengths[i] - 1, i, :])

    # output
    text_features = torch.stack(text_features)
    text_features = self.fc_output(text_features)
    return text_features

  def forward_lstm_(self, etexts):
    batch_size = etexts.shape[1]
    first_hidden = (torch.zeros(1, batch_size, self.lstm_hidden_dim),
                    torch.zeros(1, batch_size, self.lstm_hidden_dim))
    first_hidden = (first_hidden[0], first_hidden[1])
    lstm_output, last_hidden = self.lstm(etexts, first_hidden)
    return lstm_output, last_hidden



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"


# TODO(lujiang): put it into the third-party
# MIT License

# Copyright (c) 2018 Nam Vo

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

"""Metric learning functions.

Codes are modified from:
https://github.com/lugiavn/generalization-dml/blob/master/nams.py
"""

import numpy as np
import torch
import torchvision


def pairwise_distances(x, y=None):
  """
    Input: x is a Nxd matrix
           y is an optional Mxd matirx
    Output: dist is a NxM matrix where dist[i,j] is the square norm between
    x[i,:] and y[j,:]
            if y is not given then use 'y=x'.
    i.e. dist[i,j] = ||x[i,:]-y[j,:]||^2
    source:
    https://discuss.pytorch.org/t/efficient-distance-matrix-computation/9065/2
    """
  x_norm = (x**2).sum(1).view(-1, 1)
  if y is not None:
    y_t = torch.transpose(y, 0, 1)
    y_norm = (y**2).sum(1).view(1, -1)
  else:
    y_t = torch.transpose(x, 0, 1)
    y_norm = x_norm.view(1, -1)

  dist = x_norm + y_norm - 2.0 * torch.mm(x, y_t)
  # Ensure diagonal is zero if x=y
  # if y is None:
  #     dist = dist - torch.diag(dist.diag)
  return torch.clamp(dist, 0.0, np.inf)


class MyTripletLossFunc(torch.autograd.Function):

  def __init__(self, triplets):
    super(MyTripletLossFunc, self).__init__()
    self.triplets = triplets
    self.triplet_count = len(triplets)

  def forward(self, features):
    self.save_for_backward(features)

    self.distances = pairwise_distances(features).cpu().numpy()

    loss = 0.0
    triplet_count = 0.0
    correct_count = 0.0
    for i, j, k in self.triplets:
      w = 1.0
      triplet_count += w
      loss += w * np.log(1 +
                         np.exp(self.distances[i, j] - self.distances[i, k]))
      if self.distances[i, j] < self.distances[i, k]:
        correct_count += 1

    loss /= triplet_count
    return torch.FloatTensor((loss,))

  def backward(self, grad_output):
    features, = self.saved_tensors
    features_np = features.cpu().numpy()
    grad_features = features.clone() * 0.0
    grad_features_np = grad_features.cpu().numpy()

    for i, j, k in self.triplets:
      w = 1.0
      f = 1.0 - 1.0 / (
          1.0 + np.exp(self.distances[i, j] - self.distances[i, k]))
      grad_features_np[i, :] += w * f * (
          features_np[i, :] - features_np[j, :]) / self.triplet_count
      grad_features_np[j, :] += w * f * (
          features_np[j, :] - features_np[i, :]) / self.triplet_count
      grad_features_np[i, :] += -w * f * (
          features_np[i, :] - features_np[k, :]) / self.triplet_count
      grad_features_np[k, :] += -w * f * (
          features_np[k, :] - features_np[i, :]) / self.triplet_count

    for i in range(features_np.shape[0]):
      grad_features[i, :] = torch.from_numpy(grad_features_np[i, :])
    grad_features *= float(grad_output.data[0])
    return grad_features


class TripletLoss(torch.nn.Module):
  """Class for the triplet loss."""
  def __init__(self, pre_layer=None):
    super(TripletLoss, self).__init__()
    self.pre_layer = pre_layer

  def forward(self, x, triplets):
    if self.pre_layer is not None:
      x = self.pre_layer(x)
    loss = MyTripletLossFunc(triplets)(x)
    return loss


class NormalizationLayer(torch.nn.Module):
  """Class for normalization layer."""
  def __init__(self, normalize_scale=1.0, learn_scale=True):
    super(NormalizationLayer, self).__init__()
    self.norm_s = float(normalize_scale)
    if learn_scale:
      self.norm_s = torch.nn.Parameter(torch.FloatTensor((self.norm_s,)))

  def forward(self, x):
    features = self.norm_s * x / torch.norm(x, dim=1, keepdim=True).expand_as(x)
    return features



from . import Message
from ..request import Request
from ...proto import jina_pb2

_available_commands = dict(jina_pb2.RequestProto.ControlRequestProto.DESCRIPTOR.enum_values_by_name)

__all__ = ['ControlMessage']


class ControlMessage(Message):
    """
    Class of the protobuf message.

    :param command: Command with string content. (e.g. 'IDLE', 'TERMINATE', 'STATUS')
    :param pod_name: Name of the current pod.
    :param identity: The identity of the current pod
    :param args: Additional positional arguments.
    :param kwargs: Additional keyword arguments.
    """
    def __init__(self, command: str, pod_name: str = 'ctl', identity: str = '',
                 *args, **kwargs):
        """Set constructor method."""
        req = Request(jina_pb2.RequestProto())
        if command in _available_commands:
            req.control.command = getattr(jina_pb2.RequestProto.ControlRequestProto, command)
        else:
            raise ValueError(f'command "{command}" is not supported, must be one of {_available_commands}')
        super().__init__(None, req, pod_name=pod_name, identity=identity, *args, **kwargs)
        req.request_type = 'control'



import argparse

from ...logging import JinaLogger


class BaseRuntime:
    """A Jina Runtime is a procedure that blocks the main process once running (i.e. :meth:`run_forever`),
    therefore must be put into a separated thread/process. Any program/library/package/module that blocks the main
    process, can be formulated into a :class:`BaseRuntime` class and then be used in :class:`BasePea`.

     In the sequel, we call the main process/thread as ``M``, the process/thread blocked :class:`Runtime` as ``S``.

     In Jina, a :class:`BasePea` object is used to manage a :class:`Runtime` object's lifecycle. A :class:`BasePea`
     is a subclass of :class:`multiprocessing.Process` or :class:`threading.Thread`, it starts from ``M`` and once the
     ``S`` is spawned, it calls :class:`Runtime` methods in the following order:

        0. :meth:`__init__` in ``M``

        1. :meth:`setup` in ``S``

        2. :meth:`run_forever` in ``S``. Note that this will block ``S``, step 3 won't be
        reached until it is unblocked by :meth:`cancel`

        3. :meth:`teardown` in ``S``. Note that ``S`` is blocked by
        :meth:`run_forever`, this step won't be reached until step 2 is unblocked by :meth:`cancel`

     The :meth:`setup` and :meth:`teardown` pair together, which defines instructions that will be executed before
     and after. In subclasses, they are optional.

     The :meth:`run_forever` and :meth:`cancel` pair together, which introduces blocking to ``S`` and then
     unblocking from it. They are mandatory for all subclasses.

     Note that, there is no "exclusive" relation between :meth:`run_forever` and :meth:`teardown`, :meth:`teardown`
     is not about "cancelling", it is about "cleaning".

     Unlike other three methods that get invoked inside ``S``, the :meth:`cancel` is invoked in ``M`` to unblock ``S``.
     Therefore, :meth:`cancel` usually requires some special communication between ``M`` and ``S``, e.g.

        - Use :class:`threading.Event` or `multiprocessing.Event`, while :meth:`run_forever` polls for this event
        - Use ZMQ to send a message, while :meth:`run_forever` polls for this message
        - Use HTTP/REST to send a request, while :meth:`run_forever` listens to this request

     Note, another way to jump out from :meth:`run_forever` is raise exceptions from it. This will immediately move to
     :meth:`teardown`.

     .. note::
        Rule of thumb on exception handling: if you are not sure if you should handle exception inside
        :meth:`run_forever`, :meth:`cancel`, :meth:`setup`, :meth:`teardown`, then DO NOT catch exception in them.
        Exception is MUCH better handled by :class:`BasePea`.


     .. seealso::

        :class:`BasePea` for managing a :class:`Runtime` object's lifecycle.

     """

    def run_forever(self):
        """ Running the blocking procedure inside ``S``. Note, once this method is called,
        ``S`` is blocked.

        .. note::

            If this method raises any exception, :meth:`teardown` will be called.

        .. seealso::

            :meth:`cancel` for cancelling the forever loop.
        """
        raise NotImplementedError

    def cancel(self):
        """ Cancelling :meth:`run_forever` from ``M``. :meth:`cancel` usually requires some special communication
        between ``M`` and ``S``, e.g.

        - Use :class:`threading.Event` or `multiprocessing.Event`, while :meth:`run_forever` polls for this event
        - Use ZMQ to send a message, while :meth:`run_forever` polls for this message
        - Use HTTP/REST to send a request, while :meth:`run_forever` listens to this request

        .. seealso::

            :meth:`run_forever` for blocking the process/thread.
        """
        raise NotImplementedError

    def setup(self):
        """ Method called to prepare the runtime inside ``S``. Optional in subclasses.
        The default implementation does nothing.

        .. note::

            If this method raises any exception, then :meth:`run_forever` and :meth:`teardown` won't be called.

        .. note::

            Unlike :meth:`__init__` called in ``M``, :meth:`setup` is called inside ``S``.
        """
        pass

    def teardown(self):
        """ Method called immediately after :meth:`run_forever` is unblocked.
        You can tidy up things here.  Optional in subclasses. The default implementation does nothing.

        .. note::

            This method will only be called if the :meth:`setup` succeeds.
        """
        self.logger.close()

    def __init__(self, args: 'argparse.Namespace'):
        super().__init__()
        self.args = args
        if args.name:
            self.name = f'{args.name}/{self.__class__.__name__}'
        else:
            self.name = self.__class__.__name__
        self.logger = JinaLogger(self.name, **vars(self.args))



import argparse
from typing import Iterable, Callable

from pydantic import create_model, validator, Field, BaseConfig

from cli.export import _export_parser_args
from jina.helper import typename, random_identity, random_port


def _get_validator(field: str, choices: Iterable):
    """ Pydantic validator classmethod generator to validate fields exist in choices """

    def validate_arg_choices(v, values):
        if v not in choices:
            raise ValueError(f'Invalid value {v}({typename(v)} for field {field}. '
                             f'Valid choices are {choices}')
        return v

    validate_arg_choices.__qualname__ = 'validate_' + field
    return validator(field, allow_reuse=True)(validate_arg_choices)


def _get_pydantic_fields(parser: Callable[..., 'argparse.ArgumentParser']):
    all_options = {}
    choices_validators = {}

    for a in _export_parser_args(parser):
        if a.get('choices', None):
            choices_validators[f'validator_for_{a["name"]}'] = _get_validator(field=a['name'],
                                                                              choices=set(a['choices']))
        if a['required']:
            f = Field(default=...,
                      example=a['default'],
                      description=a['help'])
        elif a.get('default_factory', None):
            if a['default_factory'] == random_identity.__name__:
                f = Field(
                    default_factory=random_identity,
                    example=a['default'],
                    description=a['help'])
            elif a['default_factory'] == random_port.__name__:
                f = Field(
                    default_factory=random_port,
                    example=a['default'],
                    description=a['help'])
            else:
                raise ValueError(f'default_factory: {a["default_factory"]} is not recognizable for {a}')
        else:
            f = Field(default=a['default'],
                      description=a['help'])

        all_options[a['name']] = (a['type'], f)

    return all_options, choices_validators


def build_pydantic_model(model_name: str, module: str):
    class _PydanticConfig(BaseConfig):
        arbitrary_types_allowed = True

    from jina.parsers import helper
    helper._SHOW_ALL_ARGS, old_val = True, helper._SHOW_ALL_ARGS
    from jina.parsers import set_pea_parser, set_pod_parser
    from jina.parsers.flow import set_flow_parser
    if module == 'pod':
        parser = set_pod_parser
    elif module == 'pea':
        parser = set_pea_parser
    elif module == 'flow':
        parser = set_flow_parser
    else:
        raise TypeError(f'{module} is not supported')

    all_fields, field_validators = _get_pydantic_fields(parser)

    helper._SHOW_ALL_ARGS = old_val
    return create_model(model_name,
                        **all_fields,
                        __config__=_PydanticConfig,
                        __validators__=field_validators)



"""Module for hello world argparser"""
import argparse

from pkg_resources import resource_filename

from .base import set_base_parser
from .helper import add_arg_group, _SHOW_ALL_ARGS, _chf
from ..helper import random_identity


def mixin_hw_base_parser(parser):
    """Add the arguments for hello world to the parser

    :param parser: the parser configure
    """
    gp = add_arg_group(parser, title='General')
    gp.add_argument('--workdir', type=str, default=random_identity(),
                    help='The workdir for hello-world demo'
                         'all data, indices, shards and outputs will be saved there')
    gp.add_argument('--download-proxy', type=str,
                    help='The proxy when downloading sample data')


def set_hello_parser(parser=None):
    """Set the hello parser"""

    if not parser:
        parser = set_base_parser()

    spp = parser.add_subparsers(dest='hello',
                                description='use `%(prog)-8s [sub-command] --help` '
                                            'to get detailed information about each sub-command', required=True)

    set_hw_parser(
        spp.add_parser('fashion',
                       help='Start a simple end2end fashion images index & search demo. '
                            'This demo requires no extra dependencies.',
                       description='Run a fashion search demo',
                       formatter_class=_chf))

    set_hw_chatbot_parser(
        spp.add_parser('chatbot',
                       help='''
Start a simple Covid-19 chatbot. 

Remarks:

- Pytorch, transformers & FastAPI are required to run this demo. To install all dependencies, use 

    pip install "jina[chatbot]"

- The indexing could take 1~2 minute on a CPU machine.
''',
                       description='Run a chatbot QA demo',
                       formatter_class=_chf))

    set_hw_multimodal_parser(
        spp.add_parser('multimodal',
                       help='''
Start a simple multimodal document search. 

Remarks:

- Pytorch, torchvision, transformers & FastAPI are required to run this demo. To install all dependencies, use

    pip install "jina[multimodal]"

- The indexing could take 2~3 minute on a CPU machine.
- Downloading the dataset could take ~1 minute depending on your network.
''',
                       description='Run a multimodal search demo',
                       formatter_class=_chf))


def set_hw_parser(parser=None):
    """Set the hello world parser

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    mixin_hw_base_parser(parser)
    gp = add_arg_group(parser, title='Scalability')
    gp.add_argument('--shards', type=int,
                    default=2,
                    help='The number of shards when index and query')
    gp.add_argument('--parallel', type=int,
                    default=2,
                    help='The number of parallel when index and query')
    gp = add_arg_group(parser, title='Index')
    gp.add_argument('--uses-index', type=str,
                    default=resource_filename('jina', '/'.join(('resources', 'fashion', 'helloworld.flow.index.yml'))),
                    help='The yaml path of the index flow')
    gp.add_argument('--index-data-url', type=str,
                    default='http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-images-idx3-ubyte.gz',
                    help='The url of index data (should be in idx3-ubyte.gz format)')
    gp.add_argument('--index-labels-url', type=str,
                    default='http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-labels-idx1-ubyte.gz',
                    help='The url of index labels data (should be in idx3-ubyte.gz format)')
    gp.add_argument('--index-request-size', type=int,
                    default=1024,
                    help='The request size in indexing (the maximum number of documents that will be included in a '
                         'Request before sending it)')

    gp = add_arg_group(parser, title='Search')
    gp.add_argument('--uses-query', type=str,
                    default=resource_filename('jina', '/'.join(('resources', 'fashion', 'helloworld.flow.query.yml'))),
                    help='The yaml path of the query flow')
    gp.add_argument('--query-data-url', type=str,
                    default='http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-images-idx3-ubyte.gz',
                    help='The url of query data (should be in idx3-ubyte.gz format)')
    gp.add_argument('--query-labels-url', type=str,
                    default='http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-labels-idx1-ubyte.gz',
                    help='The url of query labels data (should be in idx3-ubyte.gz format)')
    gp.add_argument('--query-request-size', type=int,
                    default=32,
                    help='The request size in searching (the maximum number of documents that will be included in a '
                         'Request before sending it)')
    gp.add_argument('--num-query', type=int, default=128,
                    help='The number of queries to visualize')
    gp.add_argument('--top-k', type=int, default=50,
                    help='Top-k results to retrieve and visualize')
    return parser


def set_hw_chatbot_parser(parser=None):
    """Set the parser for the hello world chatbot

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    mixin_hw_base_parser(parser)
    parser.add_argument('--uses', type=str,
                        default=resource_filename('jina', '/'.join(('resources', 'chatbot', 'helloworld.flow.index.yml'))),
                        help='The yaml path of the index flow')
    parser.add_argument('--index-data-url', type=str,
                        default='https://static.jina.ai/chatbot/dataset.csv',
                        help='The url of index csv data')
    parser.add_argument('--demo-url', type=str,
                        default='https://static.jina.ai/chatbot/',
                        help='The url of the demo page')
    parser.add_argument('--port-expose',
                        type=int,
                        default=8080,
                        help='The port of the host exposed to the public')
    parser.add_argument('--parallel', type=int,
                        default=2,
                        help='The number of parallel when index and query')
    parser.add_argument('--unblock-query-flow', action='store_true', default=False,
                        help='Do not block the query flow' if _SHOW_ALL_ARGS else argparse.SUPPRESS)
    return parser


def set_hw_multimodal_parser(parser=None):
    """Set the parser for the hello world multimodal

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    mixin_hw_base_parser(parser)
    parser.add_argument('--uses', type=str,
                        default=resource_filename('jina', '/'.join(('resources', 'multimodal', 'flow-index.yml'))),
                        help='The yaml path of the index flow')
    parser.add_argument('--index-data-url', type=str,
                        default='https://static.jina.ai/multimodal/people-img.zip',
                        help='The url of index csv data')
    parser.add_argument('--demo-url', type=str,
                        default='https://static.jina.ai/multimodal/',
                        help='The url of the demo page')
    parser.add_argument('--port-expose',
                        type=int,
                        default=8080,
                        help='The port of the host exposed to the public')
    parser.add_argument('--unblock-query-flow', action='store_true', default=False,
                        help='Do not block the query flow' if _SHOW_ALL_ARGS else argparse.SUPPRESS)
    return parser



from jina.executors.rankers import Chunk2DocRanker


class WeightedRanker(Chunk2DocRanker):
    required_keys = {'weight'}

    def _get_score(self, match_idx, query_chunk_meta, match_chunk_meta, *args, **kwargs):
        scores = match_idx[[self.COL_QUERY_CHUNK_ID, self.COL_SCORE]]

        weight_score = 0.
        for k, v in scores:
            vv = 1 / (1 + v)
            weight_score += query_chunk_meta[k]['weight'] * vv

        return self.get_doc_id(match_idx), weight_score
